/******************************************************************************
 * program:     wp2latex                                                      *
 * function:    Decrypt encrypted WordPerfect 4.x and 5.x files               *
 * modul:       pass1C45.cc                                                   *
 * description: This modul decrypts file and call a regular wp5.x convertor.  *
 * licency:     GPL		                                              *
 ******************************************************************************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <strings.h>
#include "wp2latex.h"

#if defined(__UNIX__)||defined(__DJGPP__)
 #include <unistd.h>
#endif


class TconvertedPass1_xWP5:public TconvertedPass1
     {
public:
     virtual int Convert_first_pass(void);
     };
class TconvertedPass1_xWP4:public TconvertedPass1
     {
public:
     virtual int Convert_first_pass(void);
     };
class TconvertedPass1_xWP3:public TconvertedPass1
     {
public:
     virtual int Convert_first_pass(void);
     };


/*Register translators here*/
TconvertedPass1 *Factory_xWP5(void) {return new TconvertedPass1_xWP5;}
TconvertedPass1 *Factory_xWP4(void) {return new TconvertedPass1_xWP4;}
FFormatTranslator FormatWP5encrypted("WP5.x encrypted",Factory_xWP5);
FFormatTranslator FormatWP4encrypted("WP4.x encrypted",Factory_xWP4);

TconvertedPass1 *Factory_xWP3(void) {return new TconvertedPass1_xWP3;}
FFormatTranslator FormatWP3encrypted("WP3.x encrypted",Factory_xWP3);

/*----------Decrypting wp file--------------*/
extern string wpd_password,wpd_filename;
void StrGets(string & s);
void ObtainPassword(string & password);
char YesNoIgnore(const char *str);

const char *InputPasswordAgain="\nWarning! Password checksum does not match file checksum! Again Y/n/i?";


/*This function computes checksum for the given password*/
static WORD CheckSum(char *PassWord)
{
WORD csum;

csum=0;
if(PassWord==NULL) return(0);

while(*PassWord!=0)
	{
	if( *PassWord>='a' && *PassWord<='z')    /* convert to upper case */
			*PassWord -= 'a'-'A';

	csum = ( (csum >> 1) | ( csum << 15) ) ^ ( ((WORD)*PassWord)<<8 );
	PassWord++;
	}
return(csum);
}


static void DecryptWP45(FILE *Encrypted,FILE *Decrypted,const string & PassWord,FILE *ErrorFile)
{
char *buf;
WORD idx;                     /* index into buffer */
WORD pidx;                    /* index down password */
BYTE xmask;                   /* XOR mask */
DWORD len;
DWORD pos=ftell(Encrypted);
DWORD fsize=filesize(Encrypted);

 if(ErrorFile==NULL) ErrorFile=stderr;
 if((buf=(char *)malloc(4096))==NULL) return;


 pidx = 0;                      /* start of password */
 xmask = length(PassWord)+1;	/* start at password length+1 */

 if (Verbosing >= 1) printf(_("\nDecrypting ...\n\n"));

 while( pos<fsize )
	 {
	 if( fsize-pos > 4096 ) len = 4096;
			   else len = fsize-pos;

	 if( fread( buf, 1, len, Encrypted ) !=len )
		 {
		 fprintf(ErrorFile,_("\nError: Unexpected end of the input file."));
		 break;
		 }

	  for( idx=0; idx<len; idx++ )	/* now decrypt */
		 {
		 buf[idx] ^= PassWord()[pidx++] ^ xmask++;
		 if( pidx == length(PassWord) ) pidx = 0;
		 }

	   if( Decrypted )
		 {
		 if( fwrite( buf, 1, len, Decrypted ) !=len )
			 {
			 fprintf(ErrorFile,_("\nError:Couldn't write %lu bytes to the temporary file.\n"), (long unsigned)len);
			 break;
			 }
		 }
		 pos += len;
	 } /* while */

 if(buf!=NULL) free(buf);
}



int TconvertedPass1_xWP5::Convert_first_pass(void)
{
#ifdef DEBUG
  fprintf(log,"\n#Convert_pass1_WP5_encrypted() ");fflush(log);
#endif
WORD encryption;
string temp_filename;
DWORD DocumentStart;	/* position in file */
FILE *DecryptedFile=NULL;
BYTE SmallBuf[16];
TconvertedPass1 *cqPass1;

  cqPass1=GetConvertor("WP5.x");
  if(cqPass1==NULL)
	{
	RunError(0x202,"WP5.x");  /* unsupported module wp5 */
	return(-1);
	}

  DocumentStart=ftell(wpd);
  fseek(wpd,12,SEEK_SET);	//check for encrypted documents

  RdWORD_HiEnd(&encryption, wpd);
  if(encryption!=0)
	{
	if(CheckSum(wpd_password())!=encryption)
	   {
	   ObtainPassword(wpd_password);
	   while(CheckSum(wpd_password())!=encryption)
	      {
	      switch(YesNoIgnore(_(InputPasswordAgain)))
		 {
		 case 'Y':ObtainPassword(wpd_password);break;
		 case 'N':return(0);
		 case 'I':goto Ignore;
		 }
	      }
	   }
Ignore:
	temp_filename = copy(wpd_filename,0,
		GetExtension(wpd_filename())- wpd_filename()) + ".$$$";

	if((DecryptedFile=fopen(temp_filename,"wb"))==NULL)
			RunError(6,temp_filename);


	fseek(wpd,0l,SEEK_SET);
	fread( SmallBuf, 1, 16, wpd);
	SmallBuf[12] = SmallBuf[13] = 0;        /* write out first part of WP5.1 file */
	fwrite( SmallBuf, 1, 16, DecryptedFile );

	DecryptWP45(wpd,DecryptedFile,wpd_password,err);
	} /* ver5 */


  if(DecryptedFile!=NULL) /* Now perform normal conversion */
    {
    fclose(DecryptedFile);
    if((DecryptedFile=fopen(temp_filename,"rb"))==NULL) return(0);

    fseek(DecryptedFile,DocumentStart,SEEK_SET);
    cqPass1->InitMe(DecryptedFile, table, strip, log, err);
    cqPass1->Convert_first_pass();
    fclose(DecryptedFile);

#ifndef DEBUG
    unlink(temp_filename);
#endif		// DEBUG
    }
  delete cqPass1;
return(1);
}


int TconvertedPass1_xWP3::Convert_first_pass(void)
{
#ifdef DEBUG
  fprintf(log,"\n#Convert_pass1_WP3_encrypted() ");fflush(log);
#endif
WORD encryption;
string temp_filename;
DWORD DocumentStart;	/* position in file */
FILE *DecryptedFile=NULL;
BYTE SmallBuf[16];
TconvertedPass1 *cqPass1;

  cqPass1=GetConvertor("WP3.x");
  if(cqPass1==NULL)
	{
	RunError(0x202,"WP3.x");  /* unsupported module wp5 */
	return(-1);
	}

  DocumentStart=ftell(wpd);
  fseek(wpd,12,SEEK_SET);	//check for encrypted documents

  RdWORD_LoEnd(&encryption, wpd);
  if(encryption!=0)
	{
	if(CheckSum(wpd_password())!=encryption)
	   {
	   ObtainPassword(wpd_password);
	   while(CheckSum(wpd_password())!=encryption)
	      {
	      switch(YesNoIgnore(_(InputPasswordAgain)))
		 {
		 case 'Y':ObtainPassword(wpd_password);break;
		 case 'N':return(0);
		 case 'I':goto Ignore;
		 }
	      }
	   }
Ignore:
	temp_filename = copy(wpd_filename,0,
		GetExtension(wpd_filename())- wpd_filename()) + ".$$$";

	if((DecryptedFile=fopen(temp_filename,"wb"))==NULL)
			RunError(6,temp_filename);

		fseek(wpd,0l,SEEK_SET);
        for(unsigned i=0; i<19; i++)		// copy header+system use+application data
          {
	  fread( SmallBuf, 1, 16, wpd);
	  if(i==0)
	    SmallBuf[12] = SmallBuf[13] = 0;        /* write out first part of WP3.1 file */
	  fwrite( SmallBuf, 1, 16, DecryptedFile );
          }
	fread( SmallBuf, 1, 11, wpd);
	fwrite( SmallBuf, 1, 11, DecryptedFile ); //??? Why compressed data starts at 315 ???

	DecryptWP45(wpd,DecryptedFile,wpd_password,err);
	} /* ver5 */


  if(DecryptedFile!=NULL) /* Now perform normal conversion */
    {
    fclose(DecryptedFile);
    if((DecryptedFile=fopen(temp_filename,"rb"))==NULL) return(0);

    fseek(DecryptedFile,DocumentStart,SEEK_SET);
    cqPass1->InitMe(DecryptedFile, table, strip, log, err);
    cqPass1->Convert_first_pass();
    fclose(DecryptedFile);

#ifndef DEBUG
    unlink(temp_filename);
#endif		// DEBUG
    }
  delete cqPass1;
return(1);
}



int TconvertedPass1_xWP4::Convert_first_pass(void)
{
WORD encryption;
string temp_filename;
FILE *DecryptedFile=NULL;
TconvertedPass1 *cqPass1;

  cqPass1=GetConvertor("WP4.x");
  if(cqPass1==NULL)
	{
	RunError(0x202,"WP4.x");  /* unsupported module wp5 */
	return -1;
	}

  RdWORD_HiEnd(&encryption, wpd);
  if(encryption!=0)
	{
	if(CheckSum(wpd_password())!=encryption)
	   {
	   ObtainPassword(wpd_password);
	   while(CheckSum(wpd_password())!=encryption)
	      {
	      switch(YesNoIgnore(_(InputPasswordAgain)))
		 {
		 case 'Y':ObtainPassword(wpd_password);break;
		 case 'N':return(0);
		 case 'I':goto Ignore;
		 }
	      }
	   }
Ignore:
	temp_filename = copy(wpd_filename,0,
			     GetExtension(wpd_filename)- wpd_filename()) + ".$$$";
	if((DecryptedFile=fopen(temp_filename,"wb"))==NULL)
	RunError(6,temp_filename());

	DecryptWP45(wpd,DecryptedFile,wpd_password,err);
	} /* ver4 */

  if(DecryptedFile!=NULL) /* Now perform normal conversion */
    {
    fclose(DecryptedFile);
    if((DecryptedFile=fopen(temp_filename,"rb"))==NULL) return(0);

    cqPass1->InitMe(DecryptedFile, table, strip, log, err);
    cqPass1->Convert_first_pass();
    fclose(DecryptedFile);

#ifndef DEBUG
    unlink(temp_filename);
#endif		// DEBUG
    }
  delete cqPass1;
  return(1);
}

/*----- End of pass1c45. -----*/


